/*************************************************************/
/*  medianhs2.c Image median filtering program               */
/*                                                           */
/*  Uses a fast quick sort.                                  */
/*                                                           */
/*  First version: July 92                                   */
/*  Latest update: June 98, Stefan Spaennare                 */
/*************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/


#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

double timetic=(double)(CLOCKS_PER_SEC);

void qqsort(f,sr,sl,n)
int *f;
int *sr;
int *sl;
int n;
{
  int l,r,s,i,j;
  int temp,m;

  s=1;
  sl[s]=1;
  sr[s]=n;

  do {
     l=sl[s];
     r=sr[s];
     s--;
    do { 
     i=l;
     j=r;
     m=f[(l+r)/2];
     do {
        while (f[i] < m) {
           i++;
        }
        while (f[j] > m) {
           j--;
        }
        if (i <= j) {
           temp=f[i];
           f[i]=f[j];
           f[j]=temp;
           i++;
           j--;
        }
     } while (i <= j);
     if (i < r) {
        s++;
        sl[s]=i;
        sr[s]=r;
     }
     r=j;
    } while (l < r);
  } while (s > 0);
   
} /* qqsort */

int main(argc,argv)
int argc;
char *argv[];
{
  
   FILE *infile,*outfile;

   int i,j,i2,j2,k,l,bf,w,h;
   int xsize,ysize;
   int order,oo,s,ai,bi,aj,bj,m;

   int lim;

   int xys[3];

   short fff[3];

   short **image,**image2;
   short *mpimage,*mpimage2;

   int *v;

   int *sr,*sl;

   clock_t time1,timediff;


   if (argc<4) {
      printf("Usage: %s inimage.sss outimage.sss orderi limi\n",argv[0]);
      exit(0);
   } /* if */

  
   if ((infile = fopen(argv[1],"rb")) == NULL) {
      printf("File not found.\n");
      exit(0);
   } /* if */

   outfile=fopen(argv[2],"wb");

   order=atoi(argv[3]);
   lim=atoi(argv[4]);

   oo=(int)(log((double)(order*order))/log(2.0))+11;

   v=(int *)calloc(order*order+1,sizeof(int));
   sr=(int *)calloc(oo+1,sizeof(int));
   sl=(int *)calloc(oo+1,sizeof(int));

   fread(&xys[1],sizeof(int),2,infile);

   xsize=xys[1];
   ysize=xys[2];

 
   image=(short **)calloc(ysize,sizeof(short*));
   mpimage=(short *)calloc(xsize*ysize,sizeof(short));

   for (i=0; i<ysize; i++) {
      image[i]=mpimage;
      mpimage=mpimage+xsize;
   } /* for i */

   image2=(short **)calloc(ysize,sizeof(short*));
   mpimage2=(short *)calloc(xsize*ysize,sizeof(short));

   for (i=0; i<ysize; i++) {
      image2[i]=mpimage2;
      mpimage2=mpimage2+xsize;
   } /* for i */

   for (i=0; i<ysize; i++) {
      for (j=0; j<xsize; j++) {
          fread(&fff[1],sizeof(short),1,infile);
	  image[i][j]=fff[1];
      } /* for j */
   } /* for i */

   s=(order-1)/2;

   time1=clock();

for (i=0; i<ysize; i++) {

   ai=s;
   bi=s;
   if (i<s) {
      ai=i;
   } /* if */
   if (i>(ysize-1-s)) {
      bi=ysize-1-i;
   } /* if */

   for (j=0; j<xsize; j++) {

      aj=s;
      bj=s;
      if (j<s) {
         aj=j;
      } /* if */
      if (j>(xsize-1-s)) {
         bj=xsize-1-j;
      } /* if */

      m=1;

      for (k=-ai; k<=bi; k++) {
         for (l=-aj; l<=bj; l++) {
            v[m]=image[i+k][j+l];
            m++;
         } /* for l */
      } /* for k */

      qqsort(v,sr,sl,m-1); 

      if (abs(image[i][j]-v[m/2]) > lim) {
         image2[i][j]=v[m/2];
      } 
      else {
         image2[i][j]=image[i][j];
      } /* if */

    } /* for j */
} /* for i */

   timediff=clock()-time1;

   printf("%8.2f\n",(double)(timediff)/timetic);

   fwrite(&xys[1],sizeof(int),2,outfile);

   for (i=0; i<ysize; i++) {
      for (j=0; j<xsize; j++) {
         fff[1]=image2[i][j];
         fwrite(&fff[1],sizeof(short),1,outfile);
      } /* for j */
   } /* for i */

   fclose(infile);
   fclose(outfile);

} /* End  */
